/*
 *  $Id: dmch.h,v 1.6 1994/02/02 21:16:03 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

/* 
   These are the Chameleon-specific macros.

   This file also contains some information on the design of the Chameleon/MPI
   transport layer.
 */

/*D
      Chameleon-MPI transport (device) implementation

      This simple Chameleon implementation is meant to work with
      most message passing systems and does not rely on large amounts
      of message buffering.  Messages are sent in one or two packets, depending
      on the length.  If the message is short (< MPID_PACKET_SIZE), then
      the header and message are placed into a single message and sent
      together.  Long messages are sent without copying the message by
      sending the header in a separate packet. 

      This Chameleon implementation depends on adequate buffering by the
      message-passing system.  Because there is no asynchronous notification
      of messages in Chameleon (because most transport layers do not provide
      this important functionality), it is difficult to implement a
      packet-based system (i.e., one that divides messages into multiple 
      packets and manages all buffering).
D*/

#ifndef _DMCH_INCLUDED
#define _DMCH_INCLUDED

#include "tools.h"
#include "comm/comm.h"
#include "dmpiatom.h"

#define DEVICE_CHAMELEON
/* 
   This packet size should be selected such that
   (s + r*(n+h)) + c*n \approx (s+r*n) + s+r*h
   where s = latency, r = time to send a byte, n = total message length, 
   h = header size, and c = time to copy a byte.  This condition reduces to
   c n \approx s
   For a typical system with
   s = 50us
   c = .01us/byte
   this gives
   n = s / c = 50 us / (.01us/byte) = 5000 bytes
 */
/* #define MPID_PACKET_SIZE 4096 */
#define MPID_PACKET_SIZE 100000


/*
  The implementation reserves some message tags.

  (An optimization is to allow the use of all but a few very large tags
  for messages in the initial communicator, thus eliminating a separate
  header.  Messages in a different communicator would be sent on a reserved
  set of tags.  An alternate is to use the Chameleon tags for communicator
  types, making the message-passing system handle the queueing of messages
  by communicator

  PT2PT_TAG is the tag for short messages and the headers of long messages
  PT2PT2_TAG(src) is the tag for longer messages (by source).  This permits
  the header messages to be freely received into preallocated buffers, and
  for long messages to be received directly into user-buffers.
 */
#define MPID_PT2PT_TAG 0
#define MPID_PT2PT2_TAG(src) (1+(src))

/* Whether an operation should block or not */
typedef enum { MPID_NOTBLOCKING = 0, MPID_BLOCKING } MPID_BLOCKING_TYPE;
typedef struct {
    int len,			/* TOTAL length of message in BYTES */
        tag,			/* Message tag */
        context_id;		/* Internal communicator ID */
    char buffer[MPID_PACKET_SIZE];
				/* Minimum message size */
    } MPID_PACKET;
/* HeaderLen is just the length of the envelope of MPID_PACKET */
#define MPID_HEADER_LEN (sizeof(MPID_PACKET)-MPID_PACKET_SIZE)

/*
   Another option would be for the device handle to contain the initial
   packet.  I have NOT do this so as to keep down the size of the device
   handle (since I want relatively large packets)

   This needs to be structured so that there is just enough here for mpir.h.
 */
typedef struct {
    int           to, context_id, tag;
    int           rel_rank;              /* sender's rank in group */
    MPIR_BOOL     done;
    int           is_non_blocking;
    char          *buffer;
    ASYNCSendId_t sid;
        /* The following describes the buffer to be sent */
    void          *start;
    int           count;
    int           bytes_as_contig;
    MPIR_NODETYPE dataelement;
    } MPID_SHANDLE;

typedef struct {
    int           from, context_id, tag, len;
    MPIR_BOOL     done;
    int           is_posted;        /* set if in the posted receive queue */
    int           is_non_blocking;
    char          *temp;
    ASYNCRecvId_t rid;
        /* The following describes the buffer to be recv */
    void          *start;
    int           count;
    int           bytes_as_contig;
    MPIR_NODETYPE dataelement;
    } MPID_RHANDLE;

#define MPID_MIN(a,b) ((a) < (b) ? (a) : (b))

#define MPID_CAN_SEND_CONTIG

/* Error values come from dmpiatom.h */

/* 
   Since allocation is done by placing the device structure directly into
   the MPIR_?HANDLE, we don't need to allocate space.  We do, however, take
   this opportunity to initialize it...
 */
#define MPID_alloc_send_handle( a ) {(a)->done = NO;(a)->buffer= 0;}
#define MPID_alloc_recv_handle( a ) {(a)->done = NO;(a)->temp  = 0;}
#define MPID_free_send_handle( a )  if ((a)->buffer) {FREE((a)->buffer);}
#define MPID_free_recv_handle( a )  if ((a)->temp  ) {FREE((a)->temp);}
#ifdef FOO
#define MPID_address_send_handle( a ) &(a)
#define MPID_address_recv_handle( a ) &(a)

/*
   Access routines for the device's part of the structure
 */
#define MPID_set_send_tag( a, v )  (a)->tag    = v
#define MPID_set_send_context( a, v ) (a)->context_id   = (int)(v)
#define MPID_set_send_dest( a, v ) (a)->to     = v
#define MPID_set_send_start( a, v ) (a)->start = v
#define MPID_set_send_count( a, v ) (a)->count = v
#define MPID_set_send_dataelement( a, v ) (a)->dataelement = v
#define MPID_set_send_source_in_group( a, v ) (a)->rel_rank = v
#define MPID_set_recv_tag( a, v )  (a)->tag    = v
#define MPID_set_recv_context( a, v ) (a)->context_id   = (int)(v)
#define MPID_set_recv_source( a, v ) (a)->from = v
#define MPID_set_recv_start( a, v ) (a)->start = v
#define MPID_set_recv_count( a, v ) (a)->count = v
#define MPID_set_recv_dataelement( a, v ) (a)->dataelement = v
#endif
#define MPID_set_send_is_nonblocking( a, v ) (a)->is_non_blocking = v
#define MPID_set_recv_is_nonblocking( a, v ) (a)->is_non_blocking = v

/* Contact with the device layer is made here.  These call the
   routines to actually process a message (see dmch.c) */
#define MPID_post_send(dmpi_send_handle) \
    MPID_CH_post_send(dmpi_send_handle) 
#define MPID_post_recv(dmpi_recv_handle, is_available ) \
    MPID_CH_post_recv(dmpi_recv_handle, is_available ) 
#define MPID_complete_recv(dmpi_recv_handle) \
    MPID_CH_complete_recv(dmpi_recv_handle) 
#define MPID_complete_send(dmpi_send_handle,status) \
    MPID_CH_complete_send(dmpi_send_handle,status) 
#define MPID_check_device( blocking ) \
    MPID_CH_check_device( blocking )

#ifdef FOO
#define MPID_check_for_unexpected( source_dest,tag,context_id,found,handleptr)\
    MPIR_search_queue( &unexpected_recvs, source_dest, tag, context_id, found,\
		       1,handleptr )
#endif

extern void MPID_CH_myrank(), MPID_CH_mysize(), MPID_CH_End();
#define MPID_myrank( rank ) MPID_CH_myrank( rank )
#define MPID_mysize( size ) MPID_CH_mysize( size )
#define MPID_end()          MPID_CH_End()

#endif

#ifdef MPID_DEVICE_CODE
#include "../../include/dmpiatom.h"
#endif

