/*
 *  $Id: barrier.c,v 1.8 1994/01/19 21:45:44 doss Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#ifndef lint
static char vcid[] = "$Id: barrier.c,v 1.8 1994/01/19 21:45:44 doss Exp $";
#endif /* lint */

#include <mpi.h>
#include "coll.h"

/*@

MPI_Barrier - blocks the caller until all group members have called it; 
              the call returns at any process only after all group members
              have entered the call.

algorithm:  A tree-like or combine algorithm is used to broadcast a message 
            to all members of the communicator.  We can modifiy this to
            use "blocks" at a later time (see MPI_Bcast).

@*/
int MPI_Barrier ( comm )
MPI_Comm comm;
{
  int        rank, size, N2_prev, surfeit;
  int        d, dst, src;
  int        errno = MPI_SUCCESS;
  MPI_Status status;

  /* Check for valid arguments */
  if ( ( (comm            == (MPI_Comm)0) && (errno = MPI_ERR_COMM_NULL) ) ||
       ( (comm->comm_type == MPIR_INTER)  && (errno = MPI_ERR_COMM_INTER) ) ) 
    return (errno);

  /* Intialize communicator size */
  (void) MPI_Comm_size ( comm, &size );

  /* If there's only one member, this is trivial */
  if ( size > 1 ) {

    /* Initialize collective communicator */
    MPI_Comm  comm_coll = comm->comm_coll;

    (void) MPI_Comm_rank ( comm, &rank );
    (void) MPIR_Comm_N2_prev ( comm, &N2_prev );
    surfeit = size - N2_prev;

    /* Perform a combine-like operation */
    if ( rank < N2_prev ) {
      if( rank < surfeit ) {

	/* get the fanin letter from the upper "half" process: */
	dst = N2_prev + rank;

        MPI_Recv((void *)0,0,MPI_INT,dst,MPIR_BARRIER_TAG, comm_coll, &status);
      }

      /* combine on embedded N2_prev power-of-two processes */
      for (d = 1; d < N2_prev; d <<= 1) {
        dst = (rank ^ d);

	MPI_Sendrecv( (void *)0,0,MPI_INT,dst, MPIR_BARRIER_TAG,
		      (void *)0,0,MPI_INT,dst, MPIR_BARRIER_TAG, 
		      comm_coll, &status);
      }

      /* fanout data to nodes above N2_prev... */
      if ( rank < surfeit ) {
		dst = N2_prev + rank;
        MPI_Send( (void *)0, 0, MPI_INT, dst, MPIR_BARRIER_TAG, comm_coll);
      }
    } 
    else {
      /* fanin data to power of 2 subset */
      src = rank - N2_prev;
      MPI_Sendrecv( (void *)0, 0, MPI_INT, src, MPIR_BARRIER_TAG,
		    (void *)0, 0, MPI_INT, src, MPIR_BARRIER_TAG, 
		    comm_coll, &status);
    }
  } 
  return(errno); 
}




